"""
High-level interface for data access using Table Access Protocol (TAP)
"""


import secrets
import tempfile
from pathlib import Path
from typing import List, Union
from urllib.parse import quote, urlencode

import pandas as pd
import requests

from astromodule.io import batch_download_file, download_file, write_table
from astromodule.table import concat_tables


class TapService:
  def __init__(self, url):
    self.url = url
    self.http_client = requests.Session()

  def sync_query(self, query: str, save_path: Union[str, Path]):
    params = {
      'request': 'doQuery',
      'version': 1.0,
      'lang': 'ADQL',
      'phase': 'run',
      'format': 'csv',
      'query': query
    }
    req_url = self.url + '?' + urlencode(params)
    save_path = Path(save_path)
    save_path.parent.mkdir(parents=True, exist_ok=True)
    download_file(req_url, save_path, replace=True, http_client=self.http_client)

  def batch_sync_query(
    self,
    queries: List[str],
    save_paths: List[str] | List[Path] | str | Path,
    join_outputs: bool = False,
    workers: int = 3
  ):
    params = {
      'request': 'doQuery',
      'version': 1.0,
      'lang': 'ADQL',
      'phase': 'run',
      'format': 'csv'
    }
    urls = [self.url + '?' + urlencode({**params, 'query': q}) for q in queries]

    save_paths_aux = save_paths
    if join_outputs:
      tmp_folder = Path(tempfile.gettempdir()) / f'tap_{secrets.token_hex(3)}'
      tmp_folder.mkdir(parents=True)
      save_paths_aux = [tmp_folder / f'{i}.csv' for i in range(len(queries))]

    batch_download_file(
      urls,
      save_path=save_paths_aux,
      http_client=self.http_client,
      workers=workers
    )

    if join_outputs:
      combined_df = concat_tables(tmp_folder.glob('*.csv'), comment='#')
      write_table(combined_df, save_paths)





if __name__ == '__main__':
  tap = TapService('https://datalab.noirlab.edu/tap/sync')
  tap.sync_query('select top 10 psfdepth_g, psfdepth_r from ls_dr9.tractor where ra between 230.2939-0.0013 and 230.2939+0.0013 and dec between 29.7714-0.0013 and 29.7714+0.0013')
  tap.sync_query('select top 10 psfdepth_g, psfdepth_r from ls_dr9.tractor where ra between 230.2939-0.0013 and 230.2939+0.0013 and dec between 29.7714-0.0013 and 29.7714+0.0013')



class Question:
  def config(self, name: str):
    self.name = name
    ans_keys = [ans for ans in self.__class__.__dict__ if ans.isupper()]
    self.answers = [getattr(self, ans) for ans in ans_keys]
    for ans in self.answers:
      ans.question = self
    return self
  
  @property
  def t(self):
    return self.name



class Answer:
  def __init__(self, name: str):
    self.name = name
    self._question = None
    
  @property
  def question(self):
    return self._question
  
  @question.setter
  def question(self, q: Question):
    self._question = q
  
  @property
  def vt(self) -> str:
    return str(self)
  
  @property
  def fr(self) -> str:
    return str(self) + '_fraction'
  
  @property
  def db(self) -> str:
    return str(self) + '_debiased'
  
  def __str__(self) -> str:
    return self.question.name + '_' + self.name
  
  def __repr__(self) -> str:
    return str(self)



class Questions:
  class SmothOrFeatured(Question):
    SMOOTH = Answer('smooth')
    FEATURED_OR_DISK = Answer('featured-or-disk')
    ARTIFACT = Answer('artifact')
    
  class DiskEdgeOn(Question):
    YES = Answer('yes')
    NO = Answer('no')

  class HasSpiralArms(Question):
    YES = Answer('yes')
    NO = Answer('no')

  class Bar(Question):
    STRONG = Answer('strong')
    WEAK = Answer('weak')
    NO = Answer('no')

  class BulgeSize(Question):
    DOMINANT = Answer('dominant')
    LARGE = Answer('large')
    MODERATE = Answer('moderate')
    SMALL = Answer('small')
    NONE = Answer('none')

  class HowRounded(Question):
    ROUND = Answer('round')
    IN_BETWEEN = Answer('in-between')
    CIGAR_SHAPED = Answer('cigar-shaped')
      
  class EdgeOnBulge(Question):
    BOXY = Answer('boxy')
    NONE = Answer('none')
    ROUNDED = Answer('rounded')
      
  class SpiralWinding(Question):
    TIGHT = Answer('tight')
    MEDIUM = Answer('medium')
    LOOSE = Answer('loose')
      
  class SpiralArmCount(Question):
    ONE = Answer('1')
    TWO = Answer('2')
    THREE = Answer('3')
    FOUR = Answer('4')
    MORE_THAN_4 = Answer('more-than-4')
    CANT_TELL = Answer('cant-tell')
    
  class Merging(Question):
    NONE = Answer('none')
    MINOR_DISTURBANCE = Answer('minor-disturbance')
    MAJOR_DISTURBANCE = Answer('major-disturbance')
    MERGER = Answer('merger')
    
  class BarGZ(Question):
    YES = Answer('yes')
    NO = Answer('no')
  
  class BulgeSizeGZ(Question):
    DOMINANT = Answer('dominant')
    OBVIOUS = Answer('obvious')
    NONE = Answer('none')
    
  class HowRoundedGZ(Question):
    COMPLETELY = Answer('completely')
    IN_BETWEEN = Answer('in-between')
    CIGAR_SHAPED = Answer('cigar-shaped')
    
  class SpiralArmCountGZ(Question):
    ONE = Answer('1')
    TWO = Answer('2')
    THREE = Answer('3')
    FOUR = Answer('4')
    MORE_THAN_4 = Answer('more-than-4')
    
  class MergingGZ(Question):
    NEITHER = Answer('neither')
    TIDAL_DEBRIS = Answer('tidal-debris')
    BOTH = Answer('both')
    MERGER = Answer('merger')



class GD5:
  SMOOTH_OR_FEATURED = Questions.SmothOrFeatured().config('smooth-or-featured-dr5')
  DISK_EDGE_ON = Questions.DiskEdgeOn().config('disk-edge-on-dr5')
  HAS_SPIRAL_ARMS = Questions.HasSpiralArms().config('has-spiral-arms-dr5')
  BAR = Questions.Bar().config('bar-dr5')
  BULGE_SIZE = Questions.BulgeSize().config('bulge-size-dr5')
  HOW_ROUNDED = Questions.HowRounded().config('how-rounded-dr5')
  EDGE_ON_BULGE = Questions.EdgeOnBulge().config('edge-on-bulge-dr5')
  SPIRAL_WINDING = Questions.SpiralWinding().config('spiral-winding-dr5')
  SPIRAL_ARM_COUNT = Questions.SpiralArmCount().config('spiral-arm-count-dr5')
  MERGING = Questions.Merging().config('merging-dr5')



class GD8:
  SMOOTH_OR_FEATURED = Questions.SmothOrFeatured().config('smooth-or-featured-dr8')
  DISK_EDGE_ON = Questions.DiskEdgeOn().config('disk-edge-on-dr8')
  HAS_SPIRAL_ARMS = Questions.HasSpiralArms().config('has-spiral-arms-dr8')
  BAR = Questions.Bar().config('bar-dr8')
  BULGE_SIZE = Questions.BulgeSize().config('bulge-size-dr8')
  HOW_ROUNDED = Questions.HowRounded().config('how-rounded-dr8')
  EDGE_ON_BULGE = Questions.EdgeOnBulge().config('edge-on-bulge-dr8')
  SPIRAL_WINDING = Questions.SpiralWinding().config('spiral-winding-dr8')
  SPIRAL_ARM_COUNT = Questions.SpiralArmCount().config('spiral-arm-count-dr8')
  MERGING = Questions.Merging().config('merging-dr8')
  
  
  
class GD12:
  SMOOTH_OR_FEATURED = Questions.SmothOrFeatured().config('smooth-or-featured-dr12')
  DISK_EDGE_ON = Questions.DiskEdgeOn().config('disk-edge-on-dr12')
  HAS_SPIRAL_ARMS = Questions.HasSpiralArms().config('has-spiral-arms-dr12')
  BAR = Questions.BarGZ().config('bar-dr12')
  BULGE_SIZE = Questions.BulgeSizeGZ().config('bulge-size-dr12')
  HOW_ROUNDED = Questions.HowRoundedGZ().config('how-rounded-dr12')
  EDGE_ON_BULGE = Questions.EdgeOnBulge().config('edge-on-bulge-dr12')
  SPIRAL_WINDING = Questions.SpiralWinding().config('spiral-winding-dr12')
  SPIRAL_ARM_COUNT = Questions.SpiralArmCountGZ().config('spiral-arm-count-dr12')
  MERGING = Questions.MergingGZ().config('merging-dr12')



if __name__ == '__main__':
  print(GD12.SPIRAL_ARM_COUNT)


"""
Hyperparameter data model useful for standardizing the training loop of
machine learning applications.
"""


import logging
import re
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Sequence, Union

import optuna

L = logging.getLogger(__name__)


class HyperParameter(ABC):
  """
  This abstract class is the generic representation of a hyperparmaeter 
  and can be used on any model of Machine Learning or Deep Learning.
  This data model is designed to be a ``optuna`` wrapper,
  In this way, it is possible to implement a model training function without
  worry about how hyperparimeters are obtained.That is, it is possible
  Get the hyperparimeters automatically by the ``optuna`` library
  or predefined (manual) form with the same code.
  """
  def __init__(self):
    self._trial = None
    self.last_value = None
    self.attrs = {}
    
  @abstractmethod
  def suggest(self, trial: optuna.trial.FrozenTrial = None):
    """
    Used to recover the value of the hyperparmeter as defined by the
    concrete class.The concrete class can implement a logic based
    in Auto-ML using the ``optuna`` library or predefined values.
    The training function of the model should call this method to obtain the
    hyperparmeter value.

    Parameters
    ----------
    trial: optuna.trial.FrozenTrial, optional
      An instance of the ``optuna`` trial, by default none
    """
    pass

  def set_attr(self, key, value):
    self.attrs[key] = value

  def set_trial(self, trial: optuna.trial.FrozenTrial):
    self._trial = trial

  def to_dict(self, show_name: bool = False) -> dict:
    """
    A hyperparmeter can be defined from various attributes.
    This method expresses the hyperparmeter in the form of a python dictionary.

    Parameters
    ----------
    Show_Name: BOOL, OPTIONAL
      ``True`` if the `HyperParmeter` name should be included, in this case, a 
      pair key-value ``name: <hyperparameter-name>`` will be included, else,
      only hyperparmeter attributes will be included. By default ``False``

    Return
    -------
    dict
      Representation of the hyperparmeter in the form of python dictionary
    """
    if show_name:
      return self.attrs
    else:
      copy = self.attrs.copy()
      copy.pop('name')
      return copy

  def clear_last_value(self):
    """
    Sets ``last_value`` attribute to ``None``
    """
    self.last_value = None

  @staticmethod
  def from_dict(params: dict):
    """
    Creates a concrete instance of hyperparmeter from a dictionary of the
    python

    Parameters
    ----------
    params : dict
      A python dictionary that can be properly transformed into a concrete 
      hyperparmeter instance.

    Returns
    -------
    HyperParameter
      A concrete instance of hyperparmeter
    """
    t = params.pop('type')
    if t == 'categorical':
      E = CategoricalHyperParameter
    elif t == 'float':
      E = FloatHyperParameter
    elif t == 'int':
      E = IntHyperParameter
    elif t == 'constant':
      E = ConstantHyperParameter
    return E(**params)



class CategoricalHyperParameter(HyperParameter):
  """
  Concrete class that represents a hyperparimeter of the categorical type

  Parameters
  ----------
  name: str
    The hyperparâmeter identification name, must be unique.
  choices: Sequence[Any]
    A list-like object of categorical values
  """
  def __init__(self, name: str, choices: Sequence[Any]):
    super(CategoricalHyperParameter, self).__init__()
    self.set_attr('name', name)
    self.set_attr('choices', choices)

  def suggest(self, trial: optuna.trial.FrozenTrial = None) -> Any:
    _trial = trial or self._trial
    self.last_value = _trial.suggest_categorical(**self.attrs)
    return self.last_value



class FloatHyperParameter(HyperParameter):
  def __init__(
    self,
    name: str,
    low: float,
    high: float,
    step: float = None,
    log: bool = False
  ):
    super(FloatHyperParameter, self).__init__()
    self.set_attr('name', name)
    self.set_attr('low', low)
    self.set_attr('high', high)
    self.set_attr('step', step)
    self.set_attr('log', log)

  def suggest(self, trial: optuna.trial.FrozenTrial = None) -> float:
    _trial = trial or self._trial
    self.last_value = _trial.suggest_float(**self.attrs)
    return self.last_value



class IntHyperParameter(HyperParameter):
  def __init__(
    self,
    name: str,
    low: int,
    high: int,
    step: int = 1,
    log: bool = False
  ):
    super(IntHyperParameter, self).__init__()
    self.set_attr('name', name)
    self.set_attr('low', low)
    self.set_attr('high', high)
    self.set_attr('step', step)
    self.set_attr('log', log)

  def suggest(self, trial: optuna.trial.FrozenTrial = None) -> int:
    _trial = trial or self._trial
    self.last_value = _trial.suggest_int(**self.attrs)
    return self.last_value



class ConstantHyperParameter(HyperParameter):
  def __init__(self, name: str, value: Any):
    super(ConstantHyperParameter, self).__init__()
    self.set_attr('name', name)
    self.set_attr('value', value)
    self.last_value = value

  def suggest(self, trial: optuna.trial.FrozenTrial = None) -> Any:
    return self.attrs['value']


class HP:
  """
  A utilitary factory class that produces concretes `HyperParameter` instances
  according to its types. Used for convenience, to avoid importing and
  instantiating each type of hyperparmeter separately.

  Returns
  -------
  Hyperparameter
    A concrete instance of hyperparmeter.
  """ 
  @staticmethod
  def cat(name: str, choices: Sequence) -> CategoricalHyperParameter:
    return CategoricalHyperParameter(name, choices)


  @staticmethod
  def const(name: str, value: Any) -> ConstantHyperParameter:
    return ConstantHyperParameter(name, value)


  @staticmethod
  def num(
    name: str,
    low: Union[float, int],
    high: Union[float, int],
    step: Union[float, int] = None,
    log: bool = False,
    dtype: Union[float, int] = float
  ) -> Union[FloatHyperParameter, IntHyperParameter]:
    if dtype == float:
      return FloatHyperParameter(name, low, high, step, log)
    else:
      return IntHyperParameter(name, low, high, step, log)



class HyperParameterSet:
  """
  Represents a set of concretes `HyperParameter` instances and 
  handles the registration and access of hyperparameters.
  
  Parameters
  ----------
  *args: HyperParameter
    Any `HyperParameter` instance created by `HP` factory

  See Also
  --------
  astromodule.hp.HyperParameter
  astromodule.hp.CategoricalHyperParameter
  astromodule.hp.IntHyperParameter
  astromodule.hp.FloatHyperParameter
  astromodule.hp.ConstantHyperParameter
  astromodule.hp.HP
  """
  def __init__(self, *args: HyperParameter, verbose: bool = True):
    self.verbose = verbose
    self.hps = {}
    self._trial: optuna.trial.Trial = None

    for hp in args:
      if isinstance(hp, HyperParameter):
        name = hp.attrs['name']
        self.hps[name] = hp
      elif isinstance(hp, dict):
        pass
      
      
  @classmethod
  def from_dict(cls, d: Dict[str, Any]):
    hp_list = []
    for k, v in d.items():
      hp_list.append(ConstantHyperParameter(k, v))
    return HyperParameterSet(*hp_list)


  def concat(self, hyperparameters: Sequence[Union[dict, HyperParameter]]):
    """
    Parses a sequence of dictionaries that represents the hyperparameters

    Parameters
    ----------
    hyperparameters: array-like of dictionaries or array-like of `HyperParameter`
      The list of hyperparameters that will be added to this
      hyperparameters set
    """
    for hp in hyperparameters:
      if type(hp) == dict:
        name = hp['name']
        self.hps.update({ name: HyperParameter.from_dict(hp) })
      else:
        name = hp.attrs['name']
        self.hps.update({ name: hp })


  def get(
    self,
    name: str,
    trial: optuna.trial.FrozenTrial = None,
    default: Any = None,
    regex: bool = False,
  ) -> Any:
    """
    Get the value of a hyperparameter identified by its name.
    For hyperparameters different than ConstantHyperParameter, this method
    will use optuna's seggest api

    Parameters
    ----------
    name: str
      The hyperparamer name
    
    trial: optuna.trial.FrozenTrial
      The optuan trial instance
    
    default: Any
      Default value returned if the specified hyperparameter name wasn't found
      
    regex : bool
      When interpret ``name`` parameter as a regex and get all matches as a dict

    Returns
    -------
    Any
      The hyperparameter value

    See Also
    --------
    astromodule.hp.HyperParameter.suggest
    """
    if regex:
      reg = re.compile(name)
      matched_keys = filter(reg.match, self.hps.keys())
      return {k: self.hps[k].suggest(trial) for k in matched_keys}
    
    if not name in self.hps:
      if self.verbose:
        L.warning(f'Hyperparameter {name} not found! Returning default value: {str(default)}')
      return default

    return self.hps[name].suggest(trial)


  def set_trial(self, trial: optuna.trial.FrozenTrial):
    """
    Sets the optuna's trial for all hyperparameter in this set

    Parameters
    ----------
    trial: optuna.trial.FrozenTrial
      The trial that will be added
    """
    self._trial = trial
    for hp in self.hps.values():
      hp.set_trial(trial)


  def to_values_dict(self):
    """
    Returns a dict representation of this hyperparameters set with hp name
    as dict key and last optuna's suggested value as dict value
    """
    return { name: hp.last_value for name, hp in self.hps.items() }


  def clear_values_dict(self):
    """
    Clear ``last_value`` property of the `HyperParameter`, relevant when training
    with conditional hyperparameters
    """
    for hp in self.hps.values():
      hp.clear_last_value()
      
      
  def check_missing_hp(self, hps: Sequence[str]) -> List[str]:
    """
    Checks if all elements in a given list is in hyperparameters set

    Parameters
    ----------
    hps : Sequence[str]
      A list of hyperparameters name that must include the hyperparameters set

    Returns
    -------
    List[str]
      A list of missing hyperparameters
    """
    missing = []
    for hp in hps:
      if hp not in self:
        missing.append(hp)
    return missing
  
  
  def __contains__(self, key: str):
    """
    Checks if the hyperparameter set contains a hp identified by the given key

    Parameters
    ----------
    key : str
      The hyperparameter name

    Returns
    -------
    bool
      `True` if the set contains a hyperparameter with a given key,
      `False` otherwise.
    """
    return any(e == key for e in self.hps.keys())
  
  
  def __iadd__(self, hp: HyperParameter):
    hp.set_trial(self._trial)
    self.hps.update({hp.attrs.get('name'): hp})
    return self
  
  
  def add(self, hp: HyperParameter):
    self += hp
  
  
if __name__ == '__main__':
  hps = HyperParameterSet(
    HP.const('mlp_validation_fraction', 0.2),
    HP.const('mlp_early_stopping', True),
    HP.const('mlp_n_iter_no_change', 10),
    verbose=False,
  )
  hps += HP.const('A', 2)
  print(hps.hps)
  print(hps.get(r'mlp_*', regex=True))


import mimetypes
import smtplib
from email import encoders
from email.message import EmailMessage, MIMEPart
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import COMMASPACE, formatdate
from pathlib import Path
from typing import List


def send_email(
  sender: str, 
  recipients: List[str],
  username: str, 
  password: str, 
  message: str, 
  subject: str = '', 
  files: List[str] = [],
  use_tls: bool = True,
):
  msg = MIMEMultipart()
  msg['From'] = sender
  msg['To'] = COMMASPACE.join(recipients)
  msg['Date'] = formatdate(localtime=True)
  msg['Subject'] = subject
  msg.attach(MIMEText(message))
  
  for path in files:
    part = MIMEBase('application', 'octet-stream')
    with open(path, 'rb') as file:
      part.set_payload(file.read())
    encoders.encode_base64(part)
    part.add_header(
      'Content-Disposition',
      f'attachment; filename={Path(path).name}'
    )
    msg.attach(part)

  smtp = smtplib.SMTP('smtp.office365.com', 587)
  if use_tls:
    smtp.starttls()
  smtp.login(username, password)
  smtp.sendmail(sender, recipients, msg.as_string())
  smtp.quit()



def send_email2(
  sender: str, 
  recipients: List[str],
  username: str, 
  password: str, 
  message: str, 
  subject: str = '', 
  files: List[str] = [],
  use_tls: bool = True,
):
  msg = EmailMessage()
  msg['From'] = sender
  msg['To'] = COMMASPACE.join(recipients)
  msg['Date'] = formatdate(localtime=True)
  msg['Subject'] = subject
  msg.set_content(message)
  
  for path in files:
    with open(path, 'rb') as fp:
      file_data = fp.read()
      maintype, _, subtype = (mimetypes.guess_type(path)[0] or 'application/octet-stream').partition("/")
      msg.add_attachment(file_data, maintype=maintype, subtype=subtype, filename=path)

  smtp = smtplib.SMTP('smtp.office365.com', 587)
  if use_tls:
    smtp.starttls()
  smtp.login(username, password)
  smtp.sendmail(sender, recipients, msg.as_string())
  smtp.quit()



  
if __name__ == '__main__':
  send_email2(
    'natanael_magalhaes@msn.com', 
    ['palmiliaxonga@gmail.com'], 
    'natanael_magalhaes@msn.com', 
    'fat=fat*n;', 
    'Relatório Final', 
    'Relatório Final',
    files=['/home/natan/repos/splus-radial-search/output_heasarc/heasarc_top200.pdf']
  )


"""
This module performs requests to `splus.cloud <https://splus.cloud>`_ server 
over https

Features
--------
* Download a single rgb image colored using Lupton or Trilogy method
* Donwload a single fits image of single band
* Fast batch download of rgb images using multiprocessing
* Fast batch download of fits images using multiprocessing
* Single async database query
* Fast batch database queries using multiprocessing

Authors
-------
Natanael Magalhães Cardoso <`natanael.net <https://natanael.net>`_>
"""

import concurrent.futures
import os
import secrets
import tempfile
from datetime import datetime, timedelta
from functools import wraps
from io import BufferedIOBase, BytesIO, RawIOBase
from multiprocessing import Lock
from pathlib import Path
from time import sleep
from typing import Callable, List, Literal, Union
from urllib.parse import urljoin, urlparse

import requests
import tqdm

from astromodule.io import download_file, write_table
from astromodule.table import concat_tables

BASE_URL = 'https://splus.cloud/api/'
LOGIN_ROUTE = 'auth/login'
LUPTON_ROUTE = 'get_lupton_image/{ra}/{dec}/{size}/{r_band}/{g_band}/{b_band}/{stretch}/{Q}'
TRILOGY_ROUTE = 'get_image/{ra}/{dec}/{size}/{r_band}-{g_band}-{b_band}/{noise}/{saturation}'
FITS_ROUTE = 'get_cut/{ra}/{dec}/{size}/{band}'
PUBLIC_TAP_ROUTE = '/public-TAP/tap/async/'
PRIVATE_TAP_ROUTE = '/tap/tap/async/'
SPLUS_USER = os.environ.get('SPLUS_USER')
SPLUS_PASS = os.environ.get('SPLUS_PASS')



def update_authorization(f: Callable):
  """
  Decorator that can be placed on functions that needs S-PLUS authentication
  This decorator will update authorization token before the function call if 
  needed

  Parameters
  ----------
  f: the function that will be decorated

  Returns
  -------
  Callable
    The decorated function
  """
  @wraps(f)
  def wrapper(*args, **kwargs):
    this: SplusService = args[0]
    updated = this.update_token()
    if updated:
      this.client.headers.update({
        'Authorization': f'Token {this.token["value"]}'
      })
    return f(*args, **kwargs)
  return wrapper



class SplusService:
  """
  This service class interacts with splus.cloud server over https

  Parameters
  ----------
  username: str (optional)
    The username used in splus.cloud authentication, defaults to ``SPLUS_USER``
    environment variable
  password: str (optional)
    The password used in splus.cloud authentication, defaults to ``SPLUS_PASS``
    environment variable
  """
  _lock: Lock = Lock()


  def __init__(self, username: str = SPLUS_USER, password: str = SPLUS_PASS):
    self.credentials = {
      'username': username,
      'password': password
    }
    self.token: dict = None
    self.token_duration = timedelta(hours=1)
    self.client = requests.Session()


  def update_token(self) -> bool:
    """
    Updates splus.cloud authorization token

    Returns
    -------
    bool
      ``True`` if the token was updated and ``False`` otherwise
    """
    now = datetime.now()
    if self.token is None or self.token['timestamp'] < (now - self.token_duration):
      with SplusService._lock:
        if self.token is None or self.token['timestamp'] < (now - self.token_duration):
          resp = self.client.post(
            self._get_url(LOGIN_ROUTE),
            json=self.credentials
          )
          if resp.status_code == 200:
            if 'application/json' in resp.headers['Content-Type']:
              resp_body = resp.json()
              if 'token' in resp_body:
                self.token = {
                  'value': resp_body['token'],
                  'timestamp': datetime.now()
                }
                return True # updated
    return False # using cache


  def download_lupton_rgb(
    self,
    ra: float,
    dec: float,
    save_path: Union[str, Path],
    replace: bool = False,
    size: int = 128,
    r_band: str = 'I',
    g_band: str = 'R',
    b_band: str = 'G',
    stretch: Union[int, float] = 3,
    Q: Union[int, float] = 8
  ):
    """
    Downloads a single Lupton RGB image based on object posistion, this method
    accepts arguments to customize Lupton's parameters

    Parameters
    ----------
    ra: float
      The right ascension in degrees
    dec: float
      The declination in degrees
    save_path: str or Path
      The path where the file will be saved
    replace: bool (optional)
      This method checks if a file exists in ``save_path`` location before the
      download. If this parameters is ``True`` and a file exists in ``save_path``
      location, this method will ovewrite the existing file. If this parameter
      is ``False`` and a file exists in ``save_path`` location, this method will
      skip the download
    size: int (optional)
      The image size in pixels
    r_band: str (optional)
      The S-PLUS band that will be mapped as R channel of the RGB image
    g_band: str (optional)
      The S-PLUS band that will be mapped as G channel of the RGB image
    b_band: str (optional)
      The S-PLUS band that will be mapped as B channel of the RGB image
    stretch: int or float (optional)
      The ``stretch`` parameter of Lupton's formula
    Q: int or float (optional)
      The ``Q`` parameter of Lupton's formula
    """
    self._download_image(
      LUPTON_ROUTE,
      save_path=save_path,
      replace=replace,
      ra=str(ra),
      dec=str(dec),
      size=str(size),
      r_band=str(r_band),
      g_band=str(g_band),
      b_band=str(b_band),
      stretch=str(stretch),
      Q=str(Q)
    )


  def download_trilogy_rgb(
    self,
    ra: float,
    dec: float,
    save_path: Union[str, Path],
    replace: bool = False,
    size: int = 128,
    r_band: List[str] = ['R', 'I', 'F861', 'Z'],
    g_band: List[str] = ['G', 'F515', 'F660'],
    b_band: List[str] = ['U', 'F378', 'F395', 'F410', 'F430'],
    noise: float = 0.15,
    saturation: float = 0.15
  ):
    """
    Downloads a single Trilogy RGB image based on object posistion, this method
    accepts arguments to customize Trilogy parameters

    Parameters
    ----------
    ra: float
      The right ascension in degrees
    dec: float
      The declination in degrees
    save_path: str or Path
      The path where the file will be saved
    replace: bool (optional)
      This method checks if a file exists in ``save_path`` location before the
      download. If this parameters is ``True`` and a file exists in ``save_path``
      location, this method will ovewrite the existing file. If this parameter
      is ``False`` and a file exists in ``save_path`` location, this method will
      skip the download
    size: int (optional)
      The image size in pixels
    r_band: str (optional)
      The S-PLUS band that will be mapped as R channel of the RGB image
    g_band: str (optional)
      The S-PLUS band that will be mapped as G channel of the RGB image
    b_band: str (optional)
      The S-PLUS band that will be mapped as B channel of the RGB image
    noise: int or float (optional)
      The ``noise`` parameter of Trilogy algorithm
    saturation: int or float (optional)
      The ``saturation`` parameter of Trilogy algorithm
    """
    self._download_image(
      TRILOGY_ROUTE,
      ra=ra,
      dec=dec,
      save_path=save_path,
      replace=replace,
      size=size,
      r_band=','.join(r_band),
      g_band=','.join(g_band),
      b_band=','.join(b_band),
      noise=noise,
      saturation=saturation
    )


  def download_fits(
    self,
    ra: float,
    dec: float,
    save_path: Union[str, Path],
    replace: bool = False,
    size: int = 128,
    band: str = 'R'
  ):
    """
    Downloads a single FITS image based on object posistion

    Parameters
    ----------
    ra: float
      The right ascension in degrees
    dec: float
      The declination in degrees
    save_path: str or Path
      The path where the file will be saved
    replace: bool (optional)
      This method checks if a file exists in ``save_path`` location before the
      download. If this parameters is ``True`` and a file exists in ``save_path``
      location, this method will ovewrite the existing file. If this parameter
      is ``False`` and a file exists in ``save_path`` location, this method will
      skip the download
    size: int (optional)
      The image size in pixels
    band: str (optional)
      The S-PLUS band of the fits file
    """
    self._download_image(
      FITS_ROUTE,
      ra=ra,
      dec=dec,
      save_path=save_path,
      replace=replace,
      size=size,
      band=band
    )


  def batch_image_download(
    self,
    ra: List[float],
    dec: List[float],
    save_path: List[Union[str, Path]],
    img_type: Literal['trilogy', 'lupton', 'fits'] = 'lupton',
    workers: int = None,
    **kwargs
  ):
    """
    Downloads a single Lupton RGB image based on object posistion, this method
    accepts arguments to customize Lupton's parameters

    Parameters
    ----------
    ra: list of float
      A list of right ascension in degrees
    dec: list of float
      A list of declination in degrees
    save_path: list of str or list of Path
      A list of paths where the file will be saved
    img_type: str (optional)
      One of: ``lupton``, ``trilogy`` or ``fits``.
      Defaults to ``lupton``
    workers: int (optional)
      The  number of parallel processes that will be spawned. Default to a
      single process
    kwargs: Any (optional)
      Any optional parameter of :meth:`.download_lupton_rgb` if ``img_type='lupton'``,
      or  any optional parameter of :meth:`.download_trilogy_rgb`
      if ``img_type='trilogy'`` or any optional parameter of
      :meth:`.download_fits` if ``img_type='fits'``.
      These parameters must have the same type as the mentioned functions,
      i.e. pass a single value instead of a list of values and all images
      will be downloaded with the same parameter

    Examples
    --------
      >>> ra = [172.4, 193.9, 63.3]
      >>> dec = [0.42, 2.63, -1.24]
      >>> paths = ['fig1.png', 'fig2.png', 'fig3.png']
      >>> service.batch_image_download(ra, dec, paths, 'trilogy',
      ...                             size=256, noise=0.2, saturation=0.2)
    """
    assert len(ra) == len(dec) == len(save_path)

    if img_type == 'fits':
      download_function = self.download_fits
    elif img_type == 'lupton':
      download_function = self.download_lupton_rgb
    elif img_type == 'trilogy':
      download_function = self.download_trilogy_rgb

    download_args = [
      {'ra': _ra, 'dec': _dec, 'save_path': Path(_save_path), **kwargs}
      for _ra, _dec, _save_path in zip(ra, dec, save_path)
    ]

    self._batch_download(
      download_function=download_function,
      download_args=download_args,
      workers=workers
    )


  @update_authorization
  def query(
    self,
    sql: str,
    save_path: Union[str, Path],
    replace: bool = False,
    scope: Literal['public', 'private'] = 'public',
    table = None,
    fmt: str = 'text/csv'
  ):
    """
    Sends a single query to splus.cloud database

    Parameters
    ----------
    sql: str
      The sql query string
    save_path: str or Path
      The path where the query output will be saved
    repalace: bool (optional)
      This method checks if a file exists in ``save_path`` location before the
      download. If this parameters is ``True`` and a file exists in ``save_path``
      location, this method will ovewrite the existing file. If this parameter
      is ``False`` and a file exists in ``save_path`` location, this method will
      skip the download. Default to ``False``
    scope: str (optional)
      The splus.cloud scope. Can be ``public`` or ``private``. Use ``private`` only
      if you are assigned as collaborator. Defaults to ``public``
    fmt: str (optional)
      The mimi-type of query output. Defaults to ``text/csv``

    Examples
    --------
      >>> service.query('SELECT TOP 10 * FROM dr1.all_dr1', 'query.csv')
    """
    params = {
      'request': 'doQuery',
      'version': '1.0',
      'lang': 'ADQL',
      'phase': 'run',
      'query': sql,
      'format': fmt,
    }
    
    if scope == 'public':
      url = self._get_url(PUBLIC_TAP_ROUTE)
    else:
      url = self._get_url(PRIVATE_TAP_ROUTE)

    if table is None:
      resp = self.client.post(
        url,
        params=params,
        headers={
          'Accept': 'application/json',
          'Content-Type': 'application/x-www-form-urlencoded'
        }
      )
    else:
      params['upload'] = 'upload,param:uplTable'
      file = BytesIO()
      write_table(table, file, fmt='votable')
      resp = self.client.post(
        url,
        params=params,
        files={'uplTable': file},
        headers={
          'Accept': 'application/json',
        },
      )

    if resp.status_code == 200:
      self._track_tap_job(url=resp.url, save_path=save_path, replace=replace)


  def batch_query(
    self,
    sql: List[str],
    save_path: List[Union[str, Path]],
    join: bool = False,
    replace: bool = False,
    scope: Literal['public', 'private'] = 'public',
    fmt: str = 'text/csv',
    workers: int = None
  ):
    """
    Sends a batch of queries to splus.cloud database

    Parameters
    ----------
    sql: list of str
      The sql query string
    save_path: list of str or list of Path
      The path where the query output will be saved
    join: bool (optional)
      Join all tables
    repalace: bool (optional)
      This method checks if a file exists in ``save_path`` location before the
      download. If this parameters is ``True`` and a file exists in ``save_path``
      location, this method will ovewrite the existing file. If this parameter
      is ``False`` and a file exists in ``save_path`` location, this method will
      skip the download. Default to ``False``
    scope: str (optional)
      The splus.cloud scope. Can be ``public`` or ``private``. Use ``private`` only
      if you are assigned as collaborator. Defaults to ``public``
    fmt: str (optional)
      The mime-type of query output. Defaults to ``text/csv``
    workers: int (optional)
      The number of parallel processes that will be spawned. Defaults to 1
    """
    if join:
      final_path = Path(save_path)
      save_path = [
        Path(tempfile.gettempdir()) / f'spluslib_{secrets.token_hex(8)}.csv'
        for _ in range(len(sql))
      ]
    
    args = [
      {
        'sql': _sql,
        'save_path': _save_path,
        'replace': replace,
        'scope': scope,
        'fmt': fmt
      }
      for _sql, _save_path in zip(sql, save_path)
    ]

    self._batch_download(
      download_function=self.query,
      download_args=args,
      workers=workers
    )
    
    if join:
      df = concat_tables(save_path, comment='#')
      write_table(df, final_path)
      
      
  @update_authorization
  def download_field(
    self, 
    field: str, 
    output: str | Path | BufferedIOBase | RawIOBase,
    band: Literal['R', 'G', 'I', 'U', 'Z', 'F378', 'F395', 'F410', 'F430', 'F515', 'F660', 'F861'] = 'R', 
    weight_image: bool = False,
  ):
    payload = {
      'band': band,
      'dr': None,
      'fieldname': field,
      'weight': '1' if weight_image else 0
    }
    
    resp = self.client.post(
      url='https://splus.cloud/api/download_frame',
      json=payload
    )
    
    if isinstance(output, (str, Path)):
      output = Path(output)
      output.mkdir(parents=True, exist_ok=True)
      output.write_bytes(resp.content)
    elif isinstance(output, (BufferedIOBase, RawIOBase)):
      output.write(resp.content)
    


  def _get_url(self, route: str, params: dict = {}) -> str:
    """
    Get the full url based on params
    """
    return urljoin(BASE_URL, route.format(**params))


  @update_authorization
  def _track_tap_job(self, url: str, save_path: Union[str, Path], replace: bool):
    """
    Tracks the async query status in splus.cloud database

    Parameters
    ----------
    url: str
      The job url
    save_path: str or Path
      The path where the query output will be saved
    repalace: bool (optional)
      This method checks if a file exists in ``save_path`` location before the
      download. If this parameters is ``True`` and a file exists in ``save_path``
      location, this method will ovewrite the existing file. If this parameter
      is ``False`` and a file exists in ``save_path`` location, this method will
      skip the download. Default to ``False``
    """
    while True:
      resp = self.client.get(url, headers={'Accept': 'application/json'})

      if resp.status_code == 200:
        data = resp.json()
        destruction_time = datetime.fromisoformat(data['destruction'][:-1] + '+00:00')
        now = datetime.now(destruction_time.tzinfo)

        if data['phase'] == 'EXECUTING' and destruction_time > now:
          sleep(5)
        elif data['phase'] == 'COMPLETED':
          result_url = urlparse(data['results'][0]['href'])
          result_url = result_url._replace(netloc='splus.cloud').geturl()
          download_file(
            url=result_url,
            save_path=save_path,
            replace=replace,
            http_client=self.client
          )
          break
        elif data['phase'] == 'ERROR':
          message = data['error'].get('message', '')
          print(message)
          break
        else:
          break
      else:
        print(f'Status code: {resp.status_code}')
        break


  @update_authorization
  def _download_image(
    self,
    route: str,
    save_path: Union[str, Path],
    replace: bool = False,
    **kwargs
  ):
    """
    Generic download method for splus.cloud images
    This method performs a generic two stages download, required by splus.cloud

    Parameters
    ----------
    route: str
      The image request route
    save_path: str or Path
      The path where the file will be saved
    replace: bool (optional)
      This method checks if a file exists in ``save_path`` location before the
      download. If this parameters is ``True`` and a file exists in ``save_path``
      location, this method will ovewrite the existing file. If this parameter
      is ``False`` and a file exists in ```save_path`` location, this method will
      skip the download. Default to ``False``
    """
    if not replace and save_path.exists():
      return

    # Stage 1 request
    url = self._get_url(route, kwargs)
    resp = self.client.get(url)

    if resp.status_code == 200:
      if 'application/json' in resp.headers['Content-Type']:
        resp_body = resp.json()
        file_url = self._get_url(resp_body['filename'])
        # Stage 2 request
        download_file(
          file_url,
          save_path=save_path,
          replace=replace,
          http_client=self.client
        )


  def _batch_download(
    self,
    download_function: Callable,
    download_args: List[dict],
    workers: int = None
  ):
    """
    Generic batch download method.
    This method receives a donwload function and perform a multi-theread
    execution

    Parameters
    ----------
    download_function: function
      The download function that will spawned in multiple threads
    download_args: list of dict
      The list of parameters of `download_function`
    workers: int (optional)
      The number of parallel processes that will be spawned, defaults to ``None``
    """
    with concurrent.futures.ThreadPoolExecutor(max_workers=workers) as executor:
      futures = []
      download_function = download_function or download_file

      for i in range(len(download_args)):
        futures.append(executor.submit(
          download_function,
          **download_args[i]
        ))

      for future in tqdm.tqdm(
        concurrent.futures.as_completed(futures),
        total=len(futures),
        unit=' files'
      ):
        try:
          future.result()
        except Exception:
          pass






# class TableDr4Dual(Entity):
#   _name = 'dr4_dual'
#   RA = SplusColumn('RA')
#   DEC = SplusColumn('DEC')
#   photoz = SplusColumn('photoz')
    

# class DR4Schema(Entity):
#   _name = 'DR4'
#   dr4_dual = TableDr4Dual()

# class SplusClass(Entity):
#   _name = 'SPLUS'
#   DR4 = DR4Schema()
  

# Splus = SplusClass()
# Splus._config(Splus)

# sql = """
# SELECT {','.join(tables)} 
# FROM 
# """

# tb = Splus.DR4.dr4_dual
# print(
#   (tb.RA.between(3, 5) & tb.DEC.between(-4, -12) & (tb.photoz < 0.1)) |
#   (tb.photoz > 0.4)
# )




if __name__ == '__main__':
  s = SplusService()
  ra = [11.5933851, 11.8345742, 11.9053378, 12.1397573, 12.3036425]
  dec = [-1.0180862, -0.8710110, -0.8707459, -0.7373196, -0.4088959]
  path1 = ['1.png', '2.png', '3.png', '4.png', '5.png']
  path2 = ['01.png', '02.png', '03.png', '04.png', '05.png']
  path3 = ['1.fits', '2.fits', '3.fits', '4.fits', '5.fits']
  # s.batch_image_download(ra, dec, path1, img_type='lupton', workers=3, replace=True)
  # s.batch_image_download(ra, dec, path2, img_type='trilogy', workers=3, replace=True)
  # s.batch_image_download(ra, dec, path3, img_type='fits', workers=3, replace=True)


  sql = 'SELECT TOP 100 ID, RA, DEC FROM dr3.all_dr3 where id like \'%HYDRA%\''
  sql = """
  SELECT TOP 10 * 
  FROM "dr4_dual"."dr4_dual_r" AS r 
  JOIN TAP_UPLOAD.upload AS upl ON 1=CONTAINS(
    POINT('ICRS', r.ra, r.dec), CIRCLE('ICRS', upl.RA, upl.DEC, 0.00027)
  )
  """
  # path4 = [f'table{i}.csv' for i in range(10)]
  path4 = 'table0.csv'
  # s.query(sql, save_path=path4, replace=True, table=pd.DataFrame({'ra': ra, 'dec': dec}))


"""
Utility module
"""


from pathlib import Path
from typing import Any, Sequence

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord


def iauname(
  ra: float | np.ndarray,
  dec: float | np.ndarray
) -> str | Sequence[str]:
  """
  Receives the angular position(s) of the object(s) and returns IAU2000 name(s)

  Parameters
  ----------
  ra: float or array of float
    The right ascension of the object(s).
  dec: float or array of float
    The declination of the object(s).

  Example
  --------
  >>> iauname(187.70592, 12.39112)
  'J123049.42+122328.03'

  Returns
  -------
  str or list of str
    The formated IAU name of the object(s)
  """
  coord = SkyCoord(ra=ra*u.degree, dec=dec*u.degree, frame='icrs')
  ra_str = coord.ra.to_string(unit=u.hourangle, sep='', precision=2, pad=True)
  dec_str = coord.dec.to_string(sep='', precision=1, alwayssign=True, pad=True)
  if isinstance(ra_str, np.ndarray):
    r = [f'J{_ra_str}{_dec_str}' for _ra_str, _dec_str in zip(ra_str, dec_str)]
  else:
    r = f'J{ra_str}{dec_str}'
  return r



def iauname_path(
  iaunames: str | Sequence[str] = None,
  ra: float | Sequence[float] = None,
  dec: float | Sequence[float] = None,
  prefix: str | Path = '',
  suffix: str = '',
  flat: bool = False,
  return_str: bool = False,
) -> Path | Sequence[Path]:
  """
  Calculate the nested path for a given iauname

  Parameters
  ----------
  iaunames: str, List[str], optional
    Object iauname. The iauname or RA and DEC must be passed, if ``iaunames`` is
    ``None``, this function computes the iauname using the ``ra`` and ``dec``
    parameters
  ra: float, List[float], optional
    Object RA, used only if ``iaunames`` is ``None``
  dec: float, List[float], optional
    Object DEC, used only if ``iaunames`` is ``None``
  prefix: str, Path, optional
    Path that will be prepended at the begin of all paths
  suffix: str, optional
    Suffix that will be appended at the end of all paths
  flat: bool, optional
    Create the flat path with all files inside a same parent folder. This is
    not recomended for big datasets
  return_str: bool, optional
    Cast all paths to string before returning

  Example
  -------
  iaunames_path('J123049.42+122328.03', '.png')
  Path('J123/J123049.42+122328.03.png')

  Returns
  -------
  Path, List[Path]
    The iauname path
  """
  if iaunames is None:
    iaunames = iauname(ra, dec)

  if flat:
    mapping = lambda x:  Path(prefix) / (x + suffix)
  else:
    mapping = lambda x: Path(prefix) / x[:4] / (x + suffix)

  prep_output = lambda x: str(x) if return_str else x

  if isinstance(iaunames, str):
    return prep_output(mapping(iaunames))
  else:
    return [prep_output(mapping(x)) for x in iaunames]
  
  
  
def append_query_params(url: str, query_params: dict) -> str:
  index = url.find('?')
  new_url = ''

  query_string = '&'.join([f'{k}={v}' for (k, v) in query_params.items()])

  if index > -1:
    base = url[:index]
    rest = url[index:] + '&' + query_string
    new_url = base + rest
  else:
    new_url = url + '?' + query_string

  return new_url



def filter_dict(obj: dict, mask: Sequence[Any]):
  return {k: v for k, v in obj.items() if k in mask}


"""
Simple implementation of a linear pipeline with multiprocessing support
"""

import inspect
import tempfile
from abc import ABC, abstractmethod
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, Sequence

import graphviz

from astromodule.io import parallel_function_executor
from astromodule.utils import filter_dict


class PipelineStorage:
  """
  The storage used share resources between pipeline stage
  
  Attributes
  ----------
  storage: dict
    The data shared across pipeline stages
  """
  def __init__(self):
    self._storage = {}
    
  def write(self, key: str, value: Any):
    """
    Stores ``value`` identified by a ``key``

    Parameters
    ----------
    key : str
      The data identifier
    value : Any
      The data value
    """
    self._storage[key] = value
    
  def write_many(self, **data):
    """
    Stores ``value`` identified by a ``key`` in ``data``

    Parameters
    ----------
    **data: Any
      The data to be stored
    """
    for key, value in data.items():
      self._storage[key] = value
    
  def read(self, key: str) -> Any:
    """
    Retrieve the ``value`` identified by a ``key``

    Parameters
    ----------
    key : str
      The data identifier
    """
    return self._storage.get(key)
    
  def read_many(self, keys: Sequence[str]) -> Dict[str, Any]:
    """
    Retrieve the ``values`` identified by ``keys``

    Parameters
    ----------
    keys : Sequence[str]
      The data identifiers
    """
    return filter_dict(self._storage, keys)



class PipelineStage(ABC):
  """
  Base class for all pipeline stages
  
  Attributes
  ----------
  storage: PipelineStorage
    The pipeline storage. All resources of the pipeline are shared between 
    stages using this object. The `Pipeline` class creates this object
    during instantiation
    
  See Also
  --------
  astromodule.pipeline.Pipeline
  """
  
  name: str = 'Unnamed Stage'
  """Unique name that identifies a stage inside a pipeline"""
  requirements: Sequence[str] = []
  """
  A list of all resorces required by this pipeline stage. It's includes all
  shared data required by this stage using `get_data` or `run` arguments
  """
  products: Sequence[str] = []
  """
  A list of all resorces produced by this pipeline stage. It's includes all
  shared data produced by this stage using `set_data` 
  that can be accessed by another via `get_data` 
  or `run` arguments
  """
  
  @abstractmethod
  def run(self):
    """
    All concrete class of `PipelineStage` must implement this method.
    This method is called by `Pipeline.run` and `Pipeline.map_run` when
    executing the pipeline.
    
    .. note ::
      This method receives all needed data via
    
    See also
    --------
    astromodule.pipeline.Pipeline.run
    astromodule.pipeline.Pipeline.map_run
    """
    pass
  
  @property
  def storage(self) -> PipelineStorage:
    """
    The pipeline shared storage object

    Returns
    -------
    PipelineStorage
      The pipeline storage
    """
    return self._storage
  
  @storage.setter
  def storage(self, pipe_storage: PipelineStorage):
    """
    The pipeline shared storage object

    Parameters
    ----------
    pipe_storage : PipelineStorage
      A pipeline storage object
    """
    self._storage = pipe_storage
  
  def set_data(self, key: str, value: Any):
    """
    Stores shared data ``value`` identified by a ``key``

    Parameters
    ----------
    key : str
      The data identifier
    
    value : Any
      The data value
    """
    self.storage.write(key, value)
    
  def get_data(self, key: str) -> Path:
    """
    Retrieve shared data identified by a ``key``

    Parameters
    ----------
    key : str
      The data identifier
    """
    return self.storage.read(key)



class Pipeline:
  """
  A simple linear pipeline implementation with mapping and multiprocessing
  support.
  
  Parameters
  ----------
  *stages: PipelineStage
    The stages of the pipeline
  
  verbose: bool, optional
    The verbosity flag, by default True.
    
  Attributes
  ----------
  storage: PipelineStorage
    The object used to share resources between all pipeline stages
    
  See Also
  --------
  astromodule.pipeline.PipelineStorage
  astromodule.pipeline.PipelineStage
  """
  def __init__(
    self, 
    *stages: PipelineStage, 
    verbose: bool = True, 
    req_list: Sequence[str] = None,
  ):
    self.verbose = verbose
    self.storage = PipelineStorage()
    self.stages = [deepcopy(s) for s in stages]
    
    for stage in self.stages:
      stage.storage = self.storage
    
    if not req_list:
      req_list = self.extract_stages_requirements()
    self.set_stages_requirements(req_list)
    self._req_list = req_list
  
  
  def set_stages_requirements(self, requirements: Sequence[str]):
    """
    Sets ``requirements`` attribute for all pipeline stages

    Parameters
    ----------
    requirements : Sequence[str]
      A list-like object of same size as pipeline stages. Each element of 
      this list is another list of required resources keys.
    """
    for stage, req in zip(self.stages, requirements):
      stage.requirements = req
      
      
  def extract_stages_requirements(self) -> Sequence[str]:
    """
    Use code inspection to find required resources based in the assignature
    of `PipelineStage.run` method.

    Returns
    -------
    Sequence[str]
      A list of resources keys for each pipeline stage.
      
    See Also
    --------
    astromodule.pipeline.PipelineStage.run
    """
    all_reqs = []
    for stage in self.stages:
      reqs = list(inspect.signature(stage.run).parameters.keys())
      all_reqs.append(reqs)
    return all_reqs
  
    
  def run(self, validate: bool = True):
    """
    Validates and executes all stages of the pipeline

    Parameters
    ----------
    validate : bool, optional
      If `True`, a pipeline requirements validation will be performed using
      `validate` method. If `False`, the validation will be skiped, 
      by default True
    
    See Also
    --------
    astromodule.pipeline.Pipeline.validate
    """
    if validate and not self.validate():
      if self.verbose:
        print('Aborting pipeline execution due to validation fail')
      return 
    
    for i, stage in enumerate(self.stages, 1):
      if self.verbose:
        print(f'[{i} / {len(self.stages)}] Stage {stage.name}')
      
      kwargs = self.storage.read_many(stage.requirements)
      outputs = stage.run(**kwargs)
      if isinstance(outputs, dict):
        self.storage.write_many(**outputs)
      
      if self.verbose:
        print()
        
        
  def map_run(
    self, 
    key: str, 
    array: Sequence[Any], 
    workers: int = 2, 
    validate: bool = True
  ):
    """
    Validates and executes all pipeline steps in a similar way to the `run` 
    method, but using multiprocessing.
    This method has a similar implementation to MapReduce [#MapReduce]_, 
    in which a function is applied to all elements of a vector.
    In this case, the function is the pipeline itself and the vector 
    is specified by the ``array`` parameter.
    Thus, the pipeline is executed ``n`` times, where ``n`` is the size of 
    the ``array`` vector.
    For each execution, the pipeline creates an storage output with 
    identifier ``key`` whose value is the element of the vector that can
    be accessed via `PipelineStage.get_output`.

    Parameters
    ----------
    key : str
      The identifier for an element of ``array`` that can
      be accessed by a pipeline stage using `PipelineStage.get_output`.
    
    array : Sequence[Any]
      The data that will be mapped to pipeline
    
    workers : int, optional
      The number of parallel proccesses that will be spawned, by default 2
    
    validate : bool, optional
      If `True`, a pipeline requirements validation will be performed using
      `validate` method. If `False`, the validation will be skiped, 
      by default True
        
    See also
    --------
    astromodule.pipeline.Pipeline.run
    astromodule.pipeline.Pipeline.validate
    astromodule.pipeline.PipelineStage.get_output
    
    References
    ----------
    .. [#MapReduce] MapReduce - Wikipedia
      `<https://en.wikipedia.org/wiki/MapReduce>`_
    """
    if validate and not self.validate(ignore=[key]):
      if self.verbose:
        print('Aborting pipeline execution due to validation fail')
      return 
    
    params = [{'key': key, 'data': d} for d in array]
    parallel_function_executor(
      func=self._pipe_executor, 
      params=params, 
      workers=workers, 
      unit='jobs'
    )
      
      
  def validate(self, ignore: Sequence[str] = []) -> bool:
    """
    Validates the pipeline by checking whether all requirements for all 
    stages are satisfied
    
    Parameters
    ----------
    ignore : Sequence[str], optional
      Keys to ignore, by default []

    Returns
    -------
    bool
      `True` if all stages can retrieve the required resources (outputs
      and artifacts) correctly, `False` otherwise.
    """
    all_resources = set(ignore)
    problems = []
    for i, stage in enumerate(self.stages, 1):
      missing_req = set(stage.requirements) - all_resources
      if len(missing_req) > 0:
        problems.append({
          'stage_index': i, 
          'stage_name': stage.name, 
          'missing_req': missing_req
        })
      all_resources = all_resources.union(stage.products)
      
    if len(problems) > 0:
      print('Missing requirements:')
      for problem in problems:
        print(f'\t{problem["stage_index"]}. {problem["stage_name"]}')
        print(*[f'\t\t- {r}' for r in problem['missing_req']], sep='\n')
      return False
    return True
  
  
  def plot(self):
    """
    Plot the pipeline graph

    Returns
    -------
    graphviz
      A graphviz object containig the pipeline digraph that can be displayed
      in Jupyter Notebook
    """
    dot = graphviz.Digraph('Pipeline')
    for i, stage in enumerate(self.stages, 1):
      dot.node(str(i), f'{i}. {stage.name}')
    for i in range(1, len(self.stages)):
      dot.edge(str(i), str(i+1))
    dot.view(directory=tempfile.gettempdir(), cleanup=True)
    return dot
      
      
  def __repr__(self) -> str:
    """
    String object representation

    Returns
    -------
    str
      The object representation
    """
    p = [f'{i}. {s.name}' for i, s in enumerate(self.stages, 1)]
    p = '\n'.join(p)
    p = f'Pipeline:\n{p}'
    return p
  
  
  def __add__(self, other: PipelineStage) -> 'Pipeline':
    """
    Overcharge of the ``+`` operator that implements concatenation of
    pipelines or concatenation of a pipeline and a stage

    Parameters
    ----------
    other : Pipeline or PipelineStage
      The pipeline or stage that will be concatenated

    Returns
    -------
    Pipeline
      The resultant pipeline
    """
    if isinstance(other, PipelineStage):
      return Pipeline(*self.stages, other)
    elif isinstance(other, Pipeline):
      return Pipeline(*self.stages, *other.stages)
    
    
  def _pipe_executor(self, key: str, data: Any):
    """
    Wrapper function that create a new pipeline instance and execute it to
    ensure isolation of different pipelines in parallel execution

    Parameters
    ----------
    key : str
      The resource name (identifier) that will be accessible to pipeline stages
    data : Any
      The value of the mapped resource
    """
    pipe = Pipeline(*self.stages, verbose=False, req_list=self._req_list)
    pipe.storage.write(key, data)
    pipe.run(validate=False)
    del pipe


  
if __name__ == '__main__':
  import random
  import time
  class Stage1(PipelineStage):
    name = 'Stage 1'
    products = ['frame']
    
    def run(self, pipe = None):
      time.sleep(random.random())
      return {
        'frame': pipe * 2
      }
    
  class Stage2(PipelineStage):
    name = 'Stage 2'
    
    def run(self, frame):
      print(frame)
      time.sleep(random.random())
    
  p = Pipeline(Stage1(), Stage2())
  # p.plot()
  # print(p)
  # p.run()
  # p.validate()
  p.map_run('pipe', [1, 2, 3, 4, 5, 6], workers=2)


"""
Common table operations
"""


import re
import secrets
import subprocess
import tempfile
from dataclasses import dataclass
from io import BufferedIOBase, BytesIO, RawIOBase, StringIO, TextIOBase
from pathlib import Path
from typing import Literal, Sequence, Tuple, Union

import astropy.units as u
import numpy as np
import pandas as pd
from astropy.coordinates import SkyCoord, match_coordinates_sky
from astropy.table import Table

from astromodule.io import PathOrFile, TableLike, read_table, write_table

RA_REGEX = re.compile(r'^ra_?\d*$', re.IGNORECASE)
DEC_REGEX = re.compile(r'^dec_?\d*$', re.IGNORECASE)

def _match_regex_against_sequence(
  regex: re.Pattern, 
  columns: Sequence[str]
) -> Tuple[int, str] | None:
  for i, col in enumerate(columns):
    if regex.match(col):
      return i, col
  return None


def guess_coords_columns(
  df: pd.DataFrame,
  ra: str | None = None,
  dec: str | None = None,
) -> Tuple[str, str]:
  """
  Receives a pandas dataframe and try to guess the columns names used to
  identify the RA and DEC coordinates.

  Parameters
  ----------
  df : pd.DataFrame
    A pandas dataframe
  ra : str | None, optional
    The column name used to name the RA column. If a string is passed, this
    function will skip the RA guessing and will return the value of RA passed
    by this parameter. If the value is set to ``None``, this function will
    guess the RA column name using a pre-defined regular expression and will
    return the value of the first match found following the sequence of
    the columns.
  dec : str | None, optional
    The column name used to name the DEC column. If a string is passed, this
    function will skip the DEC guessing and will return the value of DEC passed
    by this parameter. If the value is set to ``None``, this function will
    guess the DEC column name using a pre-defined regular expression and will
    return the value of the first match found following the sequence of
    the columns.

  Returns
  -------
  Tuple[str, str]
    A tuple of RA and DEC columns guessed.

  Raises
  ------
  ValueError
    Raises a error if the RA or DEC columns cannot be found.
  """
  cols = df.columns.to_list()
  if ra is None:
    _, ra = _match_regex_against_sequence(RA_REGEX, cols)
  if dec is None:
    _, dec = _match_regex_against_sequence(DEC_REGEX, cols)
  if ra is None or dec is None:
    raise ValueError(
      "Can't guess RA or DEC columns, please, specify the columns names "
      "via `ra` and `dec` parameters"
    )
  return ra, dec



def table_knn(
  left: TableLike | PathOrFile,
  right: TableLike | PathOrFile,
  nthneighbor: int = 1,
  left_ra: str = 'ra',
  left_dec: str = 'dec',
  right_ra: str = 'ra',
  right_dec: str = 'dec',
) -> Tuple[np.ndarray, np.ndarray]:
  left_df = read_table(left)
  right_df = read_table(right)
  
  left_coords = SkyCoord(
    ra=left_df[left_ra].values,
    dec=left_df[left_dec].values, 
    unit=u.deg,
  )
  right_coords = SkyCoord(
    ra=right_df[right_ra].values,
    dec=right_df[right_dec].values,
    unit=u.deg,
  )
  
  idx, d, _ = match_coordinates_sky(
    left_coords,
    right_coords,
    nthneighbor=nthneighbor
  )

  return np.array(idx), np.array(d)


def fast_crossmatch(
  left: TableLike | PathOrFile,
  right: TableLike | PathOrFile,
  radius: float | u.Quantity = 1*u.arcsec,
  join: Literal['inner', 'left'] = 'inner',
  nthneighbor: int = 1,
  left_ra: str | None = None,
  left_dec: str | None = None,
  left_columns: Sequence[str] | None = None,
  right_ra: str | None = None,
  right_dec: str | None = None,
  right_columns: Sequence[str] | None = None,
  include_sep: bool = True,
):
  left_df = read_table(left)
  left_ra, left_dec = guess_coords_columns(left_df, left_ra, left_dec)
  right_df = read_table(right)
  right_ra, right_dec = guess_coords_columns(right_df, right_ra, right_dec)
  
  idx, d = table_knn(
    left_df, 
    right_df, 
    nthneighbor=nthneighbor, 
    left_ra=left_ra,
    left_dec=left_dec,
    right_ra=right_ra,
    right_dec=right_dec,
  )
  
  if isinstance(radius, u.Quantity):
    radius = radius.to(u.deg).value
  else:
    radius = u.Quantity(radius, unit=u.arcsec).to(u.deg).value

  mask = d < radius

  left_idx = mask.nonzero()[0]
  right_idx = idx[mask]
  
  if left_columns is not None:
    left_df = left_df[left_columns].copy()
  if right_columns is not None:
    right_df = right_df[right_columns].copy()
  
  if join == 'inner':
    left_masked_df = left_df.iloc[left_idx]
    right_masked_df = right_df.iloc[right_idx]
    match_df = left_masked_df.copy(deep=True)
    for col in right_masked_df.columns.to_list():
      if not col in match_df.columns:
        match_df[col] = right_masked_df[col].to_numpy()
        # TODO: include a flag "replace" in this method to indicate if t2 must
        # replace or not t1 columns. This implementation consider replace=False.
    if include_sep:
      match_df['xmatch_sep'] = d[mask]
  elif join == 'left':
    right_masked_df = right_df.iloc[right_idx]
    cols = [col for col in right_masked_df.columns if col not in left_df.columns]
    match_df = left_df.copy(deep=True)
    match_df.loc[left_idx, cols] = right_masked_df[cols].values
    if include_sep:
      match_df.loc[left_idx, 'xmatch_sep'] = d[mask]
  return match_df

  # left_df_masked = left_df.iloc[primary_idx]
  # right_df_masked = right_df.iloc[secondary_idx]

  # left_df_subsample = left_df_masked[left_columns].copy() \
  #   if left_columns is not None else left_df_masked.copy()
  # right_df_subsample = right_df_masked[right_columns].copy() \
  #   if right_columns.columns is not None else right_df_masked.copy()

  # for col in right_df_subsample.columns.tolist():
  #   left_df_subsample[col] = right_df_subsample[col].to_numpy()
  #   # TODO: include a flag "replace" in this method to indicate if t2 must
  #   # replace or not t1 columns. This implementation consider replace=True.

  # r = CrossMatchResult()
  # r.distance = d[mask]
  # r.primary_idx = primary_idx
  # r.secondary_idx = secondary_idx
  # r.table = df1_subsample
  


@dataclass
class DropDuplicatesResult:
  df: pd.DataFrame
  distances: np.ndarray
  n_iterations: int
  drop_count: int

  
def drop_duplicates(
  table: TableLike | PathOrFile,
  radius: float | u.Quantity = 1*u.arcsec,
  ra: str | None = None,
  dec: str | None = None,
  columns: Sequence[str] | None = None,
  max_iterations: int = 20,
) -> DropDuplicatesResult:
  if isinstance(radius, u.Quantity):
    radius = radius.to(u.deg).value
  else:
    radius = u.Quantity(radius, unit=u.arcsec).to(u.deg).value
  
  df = read_table(table)
  ra, dec = guess_coords_columns(df, ra, dec)
  df_coords = df[[ra, dec]].copy(deep=True)
  total_drop_count = 0
  drop_count = -1
  iteration = 0
  
  while (drop_count != 0 and iteration <= max_iterations):
    print(drop_count, iteration)
    idx, d = table_knn(
      df_coords, 
      df_coords, 
      left_ra=ra, 
      left_dec=dec, 
      right_ra=ra, 
      right_dec=dec, 
      nthneighbor=2
    )

    mask = d < radius
    primary_idx = mask.nonzero()[0]
    secondary_idx = idx[mask]
    removed_idx = []

    for pid, sid in zip(primary_idx, secondary_idx):
      if sid not in removed_idx:
        removed_idx.append(pid)

    del_mask = np.isin(idx, removed_idx, invert=True).nonzero()[0]
    len_copy_df = len(df_coords)
    df_coords = df_coords.iloc[del_mask].copy()
    
    drop_count = len_copy_df - len(df_coords)
    total_drop_count += drop_count
    iteration += 1
  
  d = d[del_mask]
  final_df = df.iloc[df_coords.index]
  if columns is not None:
    final_df = final_df[columns]
  
  return DropDuplicatesResult(
    df=final_df,
    distances=d,
    n_iterations=iteration,
    drop_count=total_drop_count
  )



  
def crossmatch(
  table1: TableLike | PathOrFile,
  table2: TableLike | PathOrFile,
  ra1: str | None = None,
  dec1: str | None = None,
  ra2: str | None = None,
  dec2: str | None = None,
  radius: float | u.Quantity = 1 * u.arcsec,
  join: Literal['1and2', '1or2', 'all1', 'all2', '1not2', '2not1', '1xor2'] = '1and2',
  find: Literal['all', 'best', 'best1', 'best2'] = 'best',
  fixcols: Literal['dups', 'all', 'none'] = 'dups',
  suffix1: str = '_1',
  suffix2: str = '_2',
  scorecol: str | None = 'xmatch_sep',
  fmt: Literal['fits', 'csv'] = 'fits',
) -> pd.DataFrame | None:
  """
  Performs a crossmatch between two tables using STILTS [#ST]_ as the backend.
  This function spawns a subprocess invoking the ``tmatch2`` [#tmatch2]_ 
  tool of the STILTS executable. 

  Parameters
  ----------
  table1 : TableLike | PathOrFile
    The first table that will be crossmatched. This parameter accepts a
    table-like object (pandas dataframe, astropy table), a path to a file
    represented as a string or pathlib.Path object, or a file object
    (BinaryIO, StringIO, file-descriptor, etc).
  
  table2 : TableLike | PathOrFile
    The second table that will be crossmatched. This parameter accepts a
    table-like object (pandas dataframe, astropy table), a path to a file
    represented as a string or pathlib.Path object, or a file object
    (BinaryIO, StringIO, file-descriptor, etc).
  
  ra1 : str | None, optional
    The name of the Right Ascension (RA) column in the first table. If ``None``
    is passed, this function will try to guess the RA column name based on
    predefined patterns using the function `~guess_coords_columns`, see this
    function's documentation for more details.
  
  dec1 : str | None, optional
    The name of the Declination (DEC) column in the first table. If ``None``
    is passed, this function will try to guess the RA column name based on
    predefined patterns using the function `~guess_coords_columns`, see this
    function's documentation for more details.
  
  ra2 : str | None, optional
    The name of the Right Ascension (RA) column in the second table. If ``None``
    is passed, this function will try to guess the RA column name based on
    predefined patterns using the function `~guess_coords_columns`, see this
    function's documentation for more details.
  
  dec2 : str | None, optional
    The name of the Declination (DEC) column in the second table. If ``None``
    is passed, this function will try to guess the RA column name based on
    predefined patterns using the function `~guess_coords_columns`, see this
    function's documentation for more details.
  
  radius : float | u.Quantity, optional
    The crossmatch max error radius. This function accepts a ``float`` value,
    that will be interpreted as ``arcsec`` unit, or a `~astropy.units.Quantity`
  
  join : Literal['1and2', '1or2', 'all1', 'all2', '1not2', '2not1', '1xor2'], optional
    Determines which rows are included in the output table. 
    The matching algorithm determines which of the rows from the first 
    table correspond to which rows from the second. This parameter 
    determines what to do with that information. Perhaps the most 
    obvious thing is to write out a table containing only rows which 
    correspond to a row in both of the two input tables. However, 
    you may also want to see the unmatched rows from one or both 
    input tables, or rows present in one table but unmatched in the other, 
    or other possibilities. The options are:

    * ``1and2``: An output row for each row represented in both input 
      tables (INNER JOIN)
    * ``1or2``: An output row for each row represented in either or both 
      of the input tables (FULL OUTER JOIN)
    * ``all1``: An output row for each matched or unmatched row in table 1 
      (LEFT OUTER JOIN)
    * ``all2``: An output row for each matched or unmatched row in table 2 
      (RIGHT OUTER JOIN)
    * ``1not2``: An output row only for rows which appear in the first 
      table but are not matched in the second table
    * ``2not1``: An output row only for rows which appear in the second 
      table but are not matched in the first table
    * ``1xor2``: An output row only for rows represented in one of the 
      input tables but not the other one

  find : Literal['all', 'best', 'best1', 'best2'], optional
    Determines what happens when a row in one table can be matched by more 
    than one row in the other table. The options are:

    * ``all``: All matches. Every match between the two tables is included 
      in the result. Rows from both of the input tables may appear multiple 
      times in the result.
    * ``best``: Best match, symmetric. The best pairs are selected in a 
      way which treats the two tables symmetrically. Any input row which 
      appears in one result pair is disqualified from appearing in any 
      other result pair, so each row from both input tables will appear 
      in at most one row in the result.
    * ``best1``: Best match for each Table 1 row. For each row in table 1, 
      only the best match from table 2 will appear in the result. Each row 
      from table 1 will appear a maximum of once in the result, but rows 
      from table 2 may appear multiple times.
    * ``best2``: Best match for each Table 2 row. For each row in table 2, 
      only the best match from table 1 will appear in the result. Each row 
      from table 2 will appear a maximum of once in the result, but rows 
      from table 1 may appear multiple times.

    The differences between ``best``, ``best1`` and ``best2`` are a bit subtle. 
    In cases where it's obvious which object in each table is the 
    best match for which object in the other, choosing betwen these 
    options will not affect the result. However, in crowded fields 
    (where the distance between objects within one or both tables is 
    typically similar to or smaller than the specified match radius) 
    it will make a difference. In this case one of the asymmetric 
    options (``best1`` or ``best2``) is usually more appropriate than best, 
    but you'll have to think about which of them suits your requirements. 
    The performance (time and memory usage) of the match may also differ 
    between these options, especially if one table is much bigger than 
    the other.
  
  fixcols : Literal['dups', 'all', 'none'], optional
    Determines how input columns are renamed before use in the output table. 
    The choices are:

    * ``none``: columns are not renamed
    * ``dups``: columns which would otherwise have duplicate names in the 
      output will be renamed to indicate which table they came from
    * ``all``: all columns will be renamed to indicate which table they 
      came from

    If columns are renamed, the new ones are determined by ``suffix*`` 
    parameters. 
  
  suffix1 : str, optional
    If the fixcols parameter is set so that input columns are renamed for 
    insertion into the output table, this parameter determines how the 
    renaming is done. It gives a suffix which is appended to all renamed 
    columns from table 1. 
  
  suffix2 : str, optional
    If the fixcols parameter is set so that input columns are renamed for 
    insertion into the output table, this parameter determines how the 
    renaming is done. It gives a suffix which is appended to all renamed 
    columns from table 2. 
  
  scorecol : str | None, optional
    Gives the name of a column in the output table to contain the "match score" 
    for each pairwise match. The meaning of this column is dependent on the 
    chosen ``matcher``, but it typically represents a distance of some kind 
    between the two matching points. If ``None`` is chosen, no score 
    column will be inserted in the output table. The default value of this 
    parameter depends on matcher. 
  
  fmt : Literal['fits', 'csv'], optional
    This function converts the two input tables to files to pass to 
    stilts backend. This parameter can be used to set the intermediate
    file types. Fits is faster and is the default file type.

  Returns
  -------
  pd.DataFrame | None
    The result table as a pandas dataframe  
  
  References
  ----------
  .. [#ST] STILTS - Starlink Tables Infrastructure Library Tool Set.
      `<https://www.star.bristol.ac.uk/mbt/stilts/>`_
  .. [#tmatch2] STILTS tmatch2 Documentation.
      `<https://www.star.bristol.ac.uk/mbt/stilts/sun256/tmatch2-usage.html>`_
  """
  tmpdir = Path(tempfile.gettempdir())
  token = secrets.token_hex(8)
  tb1_path = tmpdir / f'xmatch_in1_{token}.{fmt}'
  tb2_path = tmpdir / f'xmatch_in2_{token}.{fmt}'
  
  df1 = read_table(table1)
  df2 = read_table(table2)
  
  ra1, dec1 = guess_coords_columns(df1, ra1, dec1)
  ra2, dec2 = guess_coords_columns(df2, ra2, dec2)
  
  write_table(df1, tb1_path)
  write_table(df2, tb2_path)
  
  if isinstance(radius, u.Quantity):
    radius = float(radius.to(u.arcsec).value)
  else:
    radius = float(radius)
  
  cmd = [
    'stilts',
    'tmatch2',
    'matcher=sky',
    'progress=none',
    'runner=parallel',
    f'ifmt1={fmt}',
    f'ifmt2={fmt}',
    f'ofmt={fmt}',
    'omode=out',
    f'out=-',
    f'values1={ra1} {dec1}',
    f'values2={ra2} {dec2}',
    f'params={radius}',
    f'join={join}',
    f'find={find}',
    f'fixcols={fixcols}',
    f'suffix1={suffix1}',
    f'suffix2={suffix2}',
    f'scorecol={scorecol or ""}',
    f'in1={str(tb1_path.absolute())}',
    f'in2={str(tb2_path.absolute())}',
  ]
  
  result = subprocess.run(
    cmd,
    stderr=subprocess.PIPE,
    stdout=subprocess.PIPE,
    text=False,
  )
  
  tb1_path.unlink()
  tb2_path.unlink()
  error = result.stderr.decode().strip()
  if error:
    print('STILTS proccess exited with an error signal.')
    print(error)
    return None
  
  df_out = read_table(BytesIO(result.stdout), fmt=fmt)
  return df_out



def selfmatch(
  table: TableLike | PathOrFile,
  radius: float | u.Quantity,
  action: Literal['identify', 'keep0', 'keep1', 'wide2', 'wideN'] = 'keep1',
  ra: str | None = None,
  dec: str | None = None,
  fmt: Literal['fits', 'csv'] = 'fits',
) -> pd.DataFrame | None:
  """
  Performs a selfmatch in a table (crossmatch agains the same table) using 
  STILTS [#ST]_ as a backend (the same backend of TOPCAT [#TOPCAT]_). 
  This is useful for duplicates removal, groups detection, etc. 
  This function spawns a subprocess invoking the ``tmatch1`` [#tmatch1]_ 
  tool of the STILTS executable.

  Parameters
  ----------
  table : TableLike | PathOrFile
    The table that will be crossmatched. This parameter accepts a
    table-like object (pandas dataframe, astropy table), a path to a file
    represented as a `str` or `pathlib.Path` object, or a file object
    (BinaryIO, StringIO, file-descriptor, etc).
  
  radius : float | u.Quantity
    The crossmatch max error radius. This function accepts a ``float`` value,
    that will be interpreted as ``arcsec`` unit, or a `astropy.units.Quantity`
  
  action : Literal['identify', 'keep0', 'keep1', 'wide2', 'wideN'], optional
    Determines the form of the table which will be output as a result of the 
    internal match.

    * ``identify``: The output table is the same as the input table except 
      that it contains two additional columns, GroupID and GroupSize, 
      following the input columns. Each group of rows which matched is 
      assigned a unique integer, recorded in the GroupID column, and the 
      size of each group is recorded in the GroupSize column. Rows which 
      don't match any others (singles) have null values in both these columns.
    * ``keep0``: The result is a new table containing only "single" rows, 
      that is ones which don't match any other rows in the table. 
      Any other rows are thrown out.
    * ``keep1``: The result is a new table in which only one row 
      (the first in the input table order) from each group of matching 
      ones is retained. A subsequent intra-table match with the same 
      criteria would therefore show no matches.
    * ``wideN``: The result is a new "wide" table consisting of matched 
      rows in the input table stacked next to each other. Only groups of 
      exactly N rows in the input table are used to form the output table; 
      each row of the output table consists of the columns of the first 
      group member, followed by the columns of the second group member 
      and so on. The output table therefore has N times as many columns 
      as the input table. The column names in the new table have _1, _2, ... 
      appended to them to avoid duplication.

  ra : str | None, optional
    The name of the Right Ascension (RA) column. If ``None``
    is passed, this function will try to guess the RA column name based on
    predefined patterns using the function `~guess_coords_columns`, see this
    function's documentation for more details.
  
  dec : str | None, optional
    The name of the Declination (DEC) column. If ``None``
    is passed, this function will try to guess the RA column name based on
    predefined patterns using the function `~guess_coords_columns`, see this
    function's documentation for more details.
  
  fmt : Literal['fits', 'csv'], optional
    This function converts the input table to file before passing to 
    stilts backend. This parameter can be used to set the intermediate
    file type. Fits is faster and is the default file type.

  Returns
  -------
  pd.DataFrame | None
    A table of resulting selfmatch 
      
  References
  ----------
  .. [#TOPCAT] TOPCAT - Tool for OPerations on Catalogues And Tables.
      `<https://www.star.bristol.ac.uk/mbt/topcat/>`_
  .. [#ST] STILTS - Starlink Tables Infrastructure Library Tool Set.
      `<https://www.star.bristol.ac.uk/mbt/stilts/>`_
  .. [#tmatch1] STILTS tmatch1 Documentation.
      `<https://www.star.bristol.ac.uk/mbt/stilts/sun256/tmatch1-usage.html>`_
  """
  tmpdir = Path(tempfile.gettempdir())
  token = secrets.token_hex(8)
  in_path = tmpdir / f'xmatch_in_{token}.{fmt}'
  
  df = read_table(table)
  
  ra, dec = guess_coords_columns(df, ra, dec)
  
  write_table(df, in_path)
  
  # input_stream = BytesIO()
  # df = read_table(table)
  # ra, dec = guess_coords_columns(df, ra, dec)
  # write_table(df, input_stream, fmt=fmt)
  # input_stream.seek(0)
  
  if isinstance(radius, u.Quantity):
    radius = float(radius.to(u.arcsec).value)
  else:
    radius = float(radius)
    
  cmd = [
    'stilts',
    'tmatch1',
    'matcher=sky',
    'progress=none',
    'runner=parallel',
    f'params={radius}',
    f'values={ra} {dec}',
    f'action={action}',
    f'ifmt={fmt}',
    f'ofmt={fmt}',
    'omode=out',
    'out=-',
    f'in={str(in_path.absolute())}',
  ]
  
  result = subprocess.run(
    cmd,
    stderr=subprocess.PIPE,
    stdout=subprocess.PIPE,
    text=False,
  )
  
  in_path.unlink()
  error = result.stderr.decode().strip()
  if error:
    print('STILTS proccess exited with an error signal.')
    print(error)
    return None
  
  df_out = read_table(BytesIO(result.stdout), fmt=fmt)
  return df_out




def concat_tables(
  tables: Sequence[TableLike | PathOrFile],
  **kwargs
) -> pd.DataFrame:
  """
  Concatenate tables into a single one. This function concatenate over the
  table rows and is usefull to concatenate tables with same columns, although 
  there is no error in concatenating tables with non-existent columns in other.
  If a table does not have a certain column, the values will be filled with 
  ``NaN`` values. This function does not attempt to apply any type of 
  duplicate row removal.

  Parameters
  ----------
  tables : Sequence[TableLike  |  PathOrFile]
    A sequence of tables to be concatenated. This parameter accepts a
    table-like object (pandas dataframe, astropy table), a path to a file
    represented as a string or pathlib.Path object, or a file object
    (BinaryIO, StringIO, file-descriptor, etc).
  kwargs : Any
    Arguments that will be passed to `~astromodule.io.read_table` function

  Returns
  -------
  pd.DataFrame
    A dataframe of the concatenated table
  """
  dfs = [read_table(df, **kwargs) for df in tables]
  dfs = [df for df in dfs if isinstance(df, pd.DataFrame) and not df.empty]
  return pd.concat(dfs)



if __name__ == '__main__':
  df = selfmatch(
    Path(__file__).parent.parent / 'tests' / 'selection_claudia+prepared.csv',
    radius=45*u.arcmin,
    action='identify',
    fmt='parquet'
  )
  print(df)



"""
High-level interface to Legacy Survey data
"""


from io import BytesIO
from pathlib import Path
from typing import List, Sequence, Tuple, Union

import pandas as pd
import requests
from astropy import units as u
from astropy.table import Table

from astromodule.adql import AdqlColumn, AdqlDatabase, AdqlSchema, AdqlTable
from astromodule.io import (PathOrFile, TableLike, compress_fits,
                            download_file, parallel_function_executor,
                            read_table)
from astromodule.table import guess_coords_columns
from astromodule.tap import TapService
from astromodule.utils import append_query_params, iauname, iauname_path

LEGACY_RGB_URL = 'https://www.legacysurvey.org/viewer/jpeg-cutout'
LEGACY_RGB_URL_DEV = 'https://www.legacysurvey.org/viewer-dev/jpeg-cutout'
LEGACY_FITS_URL = 'https://www.legacysurvey.org/viewer/fits-cutout'
LEGACY_FITS_URL_DEV = 'https://www.legacysurvey.org/viewer-dev/fits-cutout'
LEGACY_TAP_SYNC_DR10 = 'https://datalab.noirlab.edu/tap/sync'


# mag_r | pixscale
# 14-15 | 0.45
# 15-16 | 0.35
# 16-17 | 0.3
# 17-18 | 0.2


class LS10Tractor(AdqlTable):
  release = AdqlColumn('release')
  brickid = AdqlColumn('brickid')
  brickname = AdqlColumn('brickname')
  objid = AdqlColumn('objid')
  brick_primary = AdqlColumn('brick_primary')
  maskbits = AdqlColumn('maskbits')
  fitbits = AdqlColumn('fitbits')
  type = AdqlColumn('type')
  ra = AdqlColumn('ra')
  dec = AdqlColumn('dec')
  ra_ivar	= AdqlColumn('ra_ivar')
  dec_ivar = AdqlColumn('dec_ivar')
  bx = AdqlColumn('bx')
  by = AdqlColumn('by')
  nobs_g = AdqlColumn('nobs_g') 
  nobs_r = AdqlColumn('nobs_r')
  nobs_i = AdqlColumn('nobs_i')
  nobs_z = AdqlColumn('nobs_z')
  nobs_w1	= AdqlColumn('nobs_w1')
  nobs_w2	= AdqlColumn('nobs_w2')
  nobs_w3	= AdqlColumn('nobs_w3')
  nobs_w4	= AdqlColumn('nobs_w4')
  rchisq_g = AdqlColumn('rchisq_g')
  rchisq_r = AdqlColumn('rchisq_r')
  rchisq_i = AdqlColumn('rchisq_i')
  rchisq_z = AdqlColumn('rchisq_z')
  rchisq_w1 = AdqlColumn('rchisq_w1')
  rchisq_w2 = AdqlColumn('rchisq_w2')
  rchisq_w3 = AdqlColumn('rchisq_w3')
  rchisq_w4 = AdqlColumn('rchisq_w4')
  mag_g = AdqlColumn('mag_g')
  mag_r = AdqlColumn('mag_r')
  mag_i = AdqlColumn('mag_i')
  mag_z = AdqlColumn('mag_z')
  mag_w1 = AdqlColumn('mag_w1')
  mag_w2 = AdqlColumn('mag_w2')
  mag_w3 = AdqlColumn('mag_w3')
  mag_w4 = AdqlColumn('mag_w4')

class LS10(AdqlSchema):
  tractor = LS10Tractor()

class LegacyClass(AdqlDatabase):
  DR10 = LS10()


Legacy = LegacyClass()
Legacy.config()


class LegacyService(TapService):
  def __init__(
    self,
    replace: bool = False,
    width: float = 256,
    height: float = 256,
    pixscale: float = 0.27,
    bands: str = 'grz',
    layer: str = 'ls-dr10',
    use_dev: bool = False,
    fmt: str = 'jpg',
    compress_fits: bool = False,
    compress_type: str = 'HCOMPRESS_1',
    hcomp_scale: int = 3,
    quantize_level: int = 10,
    quantize_method: int = -1,
    workers: int = 3,
  ):
    """
    Parameters
    ----------
    replace: bool, optional
      Replace file if exists in ``save_path`` location
    width: float, optional
      Stamp width.
    height: float, optional
      Stamp height.
    pixscale: float, optional
      Pixel scale of the sky.
    bands: str, optional
      Image bands
    layer: str, optional
      Legacy Survey image layer.
    use_dev: bool, optional
      Use the dev env of Legacy Cutout API
    fmt: str, optional
      File format. One of: ``jpg`` or ``fits``
    compress_fits: bool, optional
      Compress the downloaded fits stamp to ``fits.fz``
    workers: int, optional
      Maximum spawned threads when `batch_cutout` is called
    """
    super().__init__(LEGACY_TAP_SYNC_DR10)
    self.replace = replace
    self.width = width
    self.height = height
    self.pixscale = pixscale
    self.bands = bands
    self.layer = layer
    self.use_dev = use_dev
    self.workers = workers
    self.compress_fits = compress_fits
    self.compress_type = compress_type
    self.hcomp_scale = hcomp_scale
    self.quantize_level = quantize_level
    self.quantize_method = quantize_method
    self.http_client = requests.Session()


  def cutout(
    self,
    ra: float,
    dec: float,
    save_path: Path = None,
    base_path: str | Path = '',
    mag_r: float = None,
  ) -> None:
    """
    Downloads a single Legacy Survey object RGB stamp defined by RA and DEC.

    Parameters
    ----------
    ra: float
      Right ascension of the object.
    dec: float
      Declination of the object.
    save_path: pathlib.Path, optional
      Path where downloaded file will be stored.
    base_path: str, pathlib.Path, optional
      The path that will be appended at beggining of every paths if ``save_path``
      is ``None``.
    """
    if self.image_format == 'jpg':
      url = LEGACY_RGB_URL_DEV if self.use_dev else LEGACY_RGB_URL
    else:
      url = LEGACY_FITS_URL_DEV if self.use_dev else LEGACY_FITS_URL

    if save_path is None:
      save_path = iauname_path(
        iaunames=iauname(ra=ra, dec=dec),
        prefix=Path(base_path),
        suffix=f'.{self.image_format}'
      )

    pixscale = self.get_pixscale(mag_r) if mag_r is not None else self.pixscale

    image_url = append_query_params(url, {
      'ra': ra,
      'dec': dec,
      'width': self.width,
      'height': self.height,
      'pixscale': pixscale,
      'bands': self.bands,
      'layer': self.layer
    })

    content = download_file(
      url=image_url,
      save_path=save_path,
      http_client=self.http_client,
      replace=self.replace,
      return_bytes=self.compress_fits,
    )

    if self.compress_fits:
      compress_fits(
        file=BytesIO(content),
        compress_type=self.compress_type,
        hcomp_scale=self.hcomp_scale,
        quantize_level=self.quantize_level,
        quantize_method=self.quantize_method,
        ext=0,
        save_path=save_path,
        replace=self.replace,
      )


  def batch_cutout(
    self,
    ra: Sequence[float],
    dec: Sequence[float],
    save_path: Sequence[Path] = None,
    base_path: str | Path = '',
    mag_r: Sequence[float] = None,
  ) -> Tuple[List[Path], List[Path]]:
    """
    Downloads a list of objects defined by RA and DEC coordinates.

    The ``ra``, ``dec`` and ``save_path`` lists are mandatory and
    must have same length.

    Parameters
    ----------
    ra: List[float]
      The list of RA coordinates of the desired objects.
    dec: List[float]
      The list of DEC coordinates of the desired objects.
    save_path: List[Path], optional
      The list of path where files should be saved.
    base_path: str, Path, optional
      The path that will be appended at beggining of every paths if ``save_path``
      is ``None``.
    """
    if save_path is None:
      save_path = iauname_path(
        iaunames=iauname(ra=ra, dec=dec),
        prefix=Path(base_path),
        suffix=f'.{self.image_format}'
      )

    if mag_r is None:
      params = [
        {
          'ra': _ra,
          'dec': _dec,
          'save_path': _save_path,
        }
        for _ra, _dec, _save_path in zip(ra, dec, save_path)
      ]
    else:
      params = [
        {
          'ra': _ra,
          'dec': _dec,
          'save_path': _save_path,
          'mag_r': _mag_r,
        }
        for _ra, _dec, _save_path, _mag_r in zip(ra, dec, save_path, mag_r)
      ]

    parallel_function_executor(
      self.cutout,
      params=params,
      workers=self.workers,
      unit='file'
    )

    success = [p for p in save_path if p.exists()]
    error = [p for p in save_path if not p.exists()]
    return success, error


  def get_pixscale(self, mag_r: float | Sequence[float]) -> float | Sequence[float]:
    if isinstance(mag_r, list):
      return [self.get_pixscale(m) for m in mag_r]

    if   14 <= mag_r < 15: return 0.45
    elif 15 <= mag_r < 16: return 0.35
    elif 16 <= mag_r < 17: return 0.3
    elif 17 <= mag_r < 18: return 0.2
    
    
  def crossmatch(
    self, 
    table: TableLike | PathOrFile, 
    columns: Sequence[str] | Sequence[AdqlColumn], 
    radius: float | u.Quantity,
    save_path: str | Path,
    ra_col: str = None,
    dec_col: str = None,
    where: AdqlColumn = None,
    workers: int = 3,
  ):
    query_template = """
    SELECT TOP 1 {columns}, 
      POWER(ra - ({ra}), 2) + POWER(dec - ({dec}), 2) AS ls_separation
    FROM ls_dr10.tractor
    WHERE 
      ra BETWEEN {ra_min} AND {ra_max} AND 
      dec BETWEEN {dec_min} AND {dec_max} AND
      mag_r BETWEEN 10 AND 23
    ORDER BY ls_separation ASC
    """.strip()
    df = read_table(table)
    ra_col, dec_col = guess_coords_columns(df, ra_col, dec_col)
    tb = Legacy.DR10.tractor
    queries = []
    
    if isinstance(radius, u.Quantity):
      radius = radius.to(u.deg).value
    else:
      radius = (radius * u.arcsec).to(u.deg).value
      
    for i, row in df.iterrows():
      ra = row[ra_col]
      dec = row[dec_col]
      # match_cond = (
      #   tb.ra.between(ra - radius, ra + radius) & 
      #   tb.dec.between(dec - radius, dec + radius)
      # )
      # where = where & match_cond if where is not None else match_cond
      query = query_template.format(
        columns=','.join([str(c) for c in columns]), 
        ra_min=ra-radius,
        ra_max=ra+radius,
        dec_min=dec-radius,
        dec_max=dec+radius,
        ra=ra,
        dec=dec,
      )
      queries.append(query)
    self.batch_sync_query(
      queries=queries, 
      save_paths=save_path, 
      join_outputs=True, 
      workers=workers
    )
    


if __name__ == '__main__':
  # ls = LegacyService(workers=6)
  # ls.batch_download_legacy_rgb(
  #   ra=[185.1458 + dx/2 for dx in range(20)],
  #   dec=[12.8624 + dy/2 for dy in range(20)],
  #   save_path=[Path(f'test/{i}.jpg') for i in range(20)]
  # )
  
  ls = LegacyService()
  df = pd.DataFrame(({'ra':[184.4924, 184.4922], 'dec': [7.2737, 7.1862]}))
  ls.crossmatch(df, ['mag_r', 'mag_g', 'mag_i', 'mag_z'], 3, 'test.csv')
